/* zexy-setup autogenerated setup-file
 * generated by "./makesource.sh"
 * !! DO NOT MANUALLY EDIT  !!
 */

#include "z_zexy.h"

void z_zexy_setup(void)
{
	setup_0x260x260x7e(); /* 0x260x260x7e */
	setup_0x2e(); /* 0x2e */
	setup_0x3c0x7e(); /* 0x3c0x7e */
	setup_0x3d0x3d0x7e(); /* 0x3d0x3d0x7e */
	setup_0x3e0x7e(); /* 0x3e0x7e */
	setup_0x7c0x7c0x7e(); /* 0x7c0x7c0x7e */
	a2l_setup(); /* a2l */
	abs_tilde_setup(); /* abs~ */
	absgn_tilde_setup(); /* absgn~ */
	atoi_setup(); /* atoi */
	avg_tilde_setup(); /* avg~ */
	blockmirror_tilde_setup(); /* blockmirror~ */
	blockshuffle_tilde_setup(); /* blockshuffle~ */
	blockswap_tilde_setup(); /* blockswap~ */
	date_setup(); /* date */
	demultiplex_tilde_setup(); /* demultiplex~ */
	demultiplex_setup(); /* demultiplex */
	dfreq_tilde_setup(); /* dfreq~ */
	dirac_tilde_setup(); /* dirac~ */
	drip_setup(); /* drip */
	envrms_tilde_setup(); /* envrms~ */
	fifop_setup(); /* fifop */
	freadln_setup(); /* freadln */
	fwriteln_setup(); /* fwriteln */
	glue_setup(); /* glue */
	index_setup(); /* index */
	length_setup(); /* length */
	lifop_setup(); /* lifop */
	limiter_tilde_setup(); /* limiter~ */
	list2int_setup(); /* list2int */
	list2lists_setup(); /* list2lists */
	list2symbol_setup(); /* list2symbol */
	lister_setup(); /* lister */
	listfind_setup(); /* listfind */
	liststorage_setup(); /* liststorage */
	lpt_setup(); /* lpt */
	makesymbol_setup(); /* makesymbol */
	matchbox_setup(); /* matchbox */
	mavg_setup(); /* mavg */
	minmax_setup(); /* minmax */
	msgfile_setup(); /* msgfile */
	multiline_tilde_setup(); /* multiline~ */
	multiplex_tilde_setup(); /* multiplex~ */
	multiplex_setup(); /* multiplex */
	multireceive_setup(); /* multireceive */
	niagara_setup(); /* niagara */
	noish_tilde_setup(); /* noish~ */
	noisi_tilde_setup(); /* noisi~ */
	operating_system_setup(); /* operating_system */
	pack_tilde_setup(); /* pack~ */
	pack_setup(); /* pack */
	packel_setup(); /* packel */
	pdf_tilde_setup(); /* pdf~ */
	prime_setup(); /* prime */
	quantize_tilde_setup(); /* quantize~ */
	rawprint_setup(); /* rawprint */
	regex_setup(); /* regex */
	relay_setup(); /* relay */
	repack_setup(); /* repack */
	repeat_setup(); /* repeat */
	route_tilde_setup(); /* route~ */
	sfplay_setup(); /* sfplay */
	sfrecord_setup(); /* sfrecord */
	sgn_tilde_setup(); /* sgn~ */
	sigzero_tilde_setup(); /* sigzero~ */
	sleepgrain_setup(); /* sleepgrain */
	sort_setup(); /* sort */
	step_tilde_setup(); /* step~ */
	strcmp_setup(); /* strcmp */
	sum_setup(); /* sum */
	swap_tilde_setup(); /* swap~ */
	symbol2list_setup(); /* symbol2list */
	tabdump_setup(); /* tabdump */
	tabminmax_setup(); /* tabminmax */
	tabread4_tilde_tilde_setup(); /* tabread4~~ */
	tabset_setup(); /* tabset */
	tavg_tilde_setup(); /* tavg~ */
	time_setup(); /* time */
	unpack_tilde_setup(); /* unpack~ */
	unpack_setup(); /* unpack */
	urn_setup(); /* urn */
	wrap_setup(); /* wrap */
	z_tilde_setup(); /* z~ */
}

