/* zexy-setup autogenerated header-file
 * generated by "./makesource.sh"
 * !! DO NOT MANUALLY EDIT  !!
 */

#ifndef Z_ZEXY_H__
#define Z_ZEXY_H__
void setup_0x260x260x7e(void); /* 0x260x260x7e */
void setup_0x2e(void); /* 0x2e */
void setup_0x3c0x7e(void); /* 0x3c0x7e */
void setup_0x3d0x3d0x7e(void); /* 0x3d0x3d0x7e */
void setup_0x3e0x7e(void); /* 0x3e0x7e */
void setup_0x7c0x7c0x7e(void); /* 0x7c0x7c0x7e */
void a2l_setup(void); /* a2l */
void abs_tilde_setup(void); /* abs~ */
void absgn_tilde_setup(void); /* absgn~ */
void atoi_setup(void); /* atoi */
void avg_tilde_setup(void); /* avg~ */
void blockmirror_tilde_setup(void); /* blockmirror~ */
void blockshuffle_tilde_setup(void); /* blockshuffle~ */
void blockswap_tilde_setup(void); /* blockswap~ */
void date_setup(void); /* date */
void demultiplex_tilde_setup(void); /* demultiplex~ */
void demultiplex_setup(void); /* demultiplex */
void dfreq_tilde_setup(void); /* dfreq~ */
void dirac_tilde_setup(void); /* dirac~ */
void drip_setup(void); /* drip */
void envrms_tilde_setup(void); /* envrms~ */
void fifop_setup(void); /* fifop */
void freadln_setup(void); /* freadln */
void fwriteln_setup(void); /* fwriteln */
void glue_setup(void); /* glue */
void index_setup(void); /* index */
void length_setup(void); /* length */
void lifop_setup(void); /* lifop */
void limiter_tilde_setup(void); /* limiter~ */
void list2int_setup(void); /* list2int */
void list2lists_setup(void); /* list2lists */
void list2symbol_setup(void); /* list2symbol */
void lister_setup(void); /* lister */
void listfind_setup(void); /* listfind */
void liststorage_setup(void); /* liststorage */
void lpt_setup(void); /* lpt */
void makesymbol_setup(void); /* makesymbol */
void matchbox_setup(void); /* matchbox */
void mavg_setup(void); /* mavg */
void minmax_setup(void); /* minmax */
void msgfile_setup(void); /* msgfile */
void multiline_tilde_setup(void); /* multiline~ */
void multiplex_tilde_setup(void); /* multiplex~ */
void multiplex_setup(void); /* multiplex */
void multireceive_setup(void); /* multireceive */
void niagara_setup(void); /* niagara */
void noish_tilde_setup(void); /* noish~ */
void noisi_tilde_setup(void); /* noisi~ */
void operating_system_setup(void); /* operating_system */
void pack_tilde_setup(void); /* pack~ */
void pack_setup(void); /* pack */
void packel_setup(void); /* packel */
void pdf_tilde_setup(void); /* pdf~ */
void prime_setup(void); /* prime */
void quantize_tilde_setup(void); /* quantize~ */
void rawprint_setup(void); /* rawprint */
void regex_setup(void); /* regex */
void relay_setup(void); /* relay */
void repack_setup(void); /* repack */
void repeat_setup(void); /* repeat */
void route_tilde_setup(void); /* route~ */
void sfplay_setup(void); /* sfplay */
void sfrecord_setup(void); /* sfrecord */
void sgn_tilde_setup(void); /* sgn~ */
void sigzero_tilde_setup(void); /* sigzero~ */
void sleepgrain_setup(void); /* sleepgrain */
void sort_setup(void); /* sort */
void step_tilde_setup(void); /* step~ */
void strcmp_setup(void); /* strcmp */
void sum_setup(void); /* sum */
void swap_tilde_setup(void); /* swap~ */
void symbol2list_setup(void); /* symbol2list */
void tabdump_setup(void); /* tabdump */
void tabminmax_setup(void); /* tabminmax */
void tabread4_tilde_tilde_setup(void); /* tabread4~~ */
void tabset_setup(void); /* tabset */
void tavg_tilde_setup(void); /* tavg~ */
void time_setup(void); /* time */
void unpack_tilde_setup(void); /* unpack~ */
void unpack_setup(void); /* unpack */
void urn_setup(void); /* urn */
void wrap_setup(void); /* wrap */
void z_tilde_setup(void); /* z~ */

void z_zexy_setup(void);
#endif /* Z_ZEXY_H__ */

